
#define  norm_opt_none     1
#define  norm_opt_destarea 2
#define  norm_opt_fracarea 3

#define  map_type_conserv  1
#define  map_type_bilinear 2
#define  map_type_bicubic  3
#define  map_type_distwgt  4
#define  map_type_distwgt1 5

#define  restrict_latitude 1
#define  restrict_latlon   2

/* constants */

#define  zero     0.0
#define  one      1.0
#define  two      2.0
#define  three    3.0
#define  four     4.0
#define  five     5.0
#define  half     0.5
#define  quart    0.25
#define  bignum   1.e+20
#define  tiny     1.e-14
#define  pi       3.14159265359
#define  pi2      two*pi
#define  pih      half*pi

#define  deg2rad  (pi/180.)   /* conversion for deg to rad */
#define  rad2deg  (180./pi)   /* conversion for rad to deg */


typedef struct {
  int      pinit;            /* TRUE if the pointers are initialized     */
  int      gridID1;
  int      gridID2;
  int      grid1_size, grid2_size; /* total points on each grid */
  int      grid1_rank, grid2_rank; /* rank of each grid */
  int      grid1_corners, grid2_corners; /* number of corners for each grid cell */

  int      grid1_dims[2], grid2_dims[2]; /* size of each grid dimension */
 
  int     *grid1_mask;         /* flag which cells participate */
  int     *grid2_mask;         /* flag which cells participate */

  double  *grid1_center_lat;   /* lat/lon coordinates for      */
  double  *grid1_center_lon;   /* each grid center in radians  */
  double  *grid2_center_lat; 
  double  *grid2_center_lon;
  double  *grid1_area;         /* tot area of each grid1 cell     */
  double  *grid2_area;         /* tot area of each grid2 cell     */
  /* double  *grid1_area_in; */     /* area of grid1 cell from file    */
  /* double  *grid2_area_in; */     /* area of grid2 cell from file    */
  double  *grid1_frac;         /* fractional area of grid cells   */
  double  *grid2_frac;         /* participating in remapping      */

  double  *grid1_corner_lat;   /* lat/lon coordinates for         */
  double  *grid1_corner_lon;   /* each grid corner in radians     */
  double  *grid2_corner_lat; 
  double  *grid2_corner_lon;

  int      luse_grid_centers;  /* use centers for bounding boxes  */
  /* int      luse_grid1_area;   */ /* use area from grid file         */
  /* int      luse_grid2_area;   */ /* use area from grid file         */

  double  *grid1_bound_box;    /* lat/lon bounding box for use    */
  double  *grid2_bound_box;    /* in restricting grid searches    */

  int      restrict_type;
  int      num_srch_bins;      /* num of bins for restricted srch */

  int     *bin_addr1;       /* min,max adds for grid1 cells in this lat bin  */
  int     *bin_addr2;       /* min,max adds for grid2 cells in this lat bin  */

  double  *bin_lats;        /* min,max latitude for each search bin   */
  double  *bin_lons;        /* min,max longitude for each search bin  */
}
REMAPGRID;

typedef struct {
  int   pinit;            /* TRUE if the pointers are initialized     */
  int   max_links;        /* current size of link arrays              */
  int   num_links;        /* actual number of links for remapping     */
  int   num_wts;          /* num of weights used in remapping         */
  int   map_type;         /* identifier for remapping method          */
  int   norm_opt;         /* option for normalization (conserv only)  */
  int   resize_increment; /* default amount to increase array size    */

  int  *grid1_add;        /* grid1 address for each link in mapping 1 */
  int  *grid2_add;        /* grid2 address for each link in mapping 1 */

  double *wts[4];         /* map weights for each link [max_links][num_wts] */
}
REMAPVARS;

void remapGridInit(int map_type, int gridID1, int gridID2, REMAPGRID *rg);
void remapVarsInit(int map_type, REMAPGRID *rg, REMAPVARS *rv);

void remapVarsFree(REMAPVARS *rv);
void remapGridFree(REMAPGRID *rg);

void remap(double *dst_array, double missval, int dst_size, int dst_array_dim,
	   double **map_wts, int map_wts_dim,
	   int *dst_add, int *src_add, 
	   double *src_array, double *src_grad1, double *src_grad2, double *src_grad3);

void remap_con1(double *dst_array, double missval, int dst_size, int num_links, double **map_wts,
		int num_wts, int *dst_add, int *src_add, double *src_array);

void remap_bilin(REMAPGRID *rg, REMAPVARS *rv);
void remap_bicub(REMAPGRID *rg, REMAPVARS *rv);
void remap_conserv(REMAPGRID *rg, REMAPVARS *rv);
void remap_distwgt(REMAPGRID *rg, REMAPVARS *rv);
void remap_distwgt1(REMAPGRID *rg, REMAPVARS *rv);

void resize_remap_vars(REMAPVARS *rv, int increment);

void remap_stat(REMAPGRID rg, REMAPVARS rv, double *array1, double *array2, double missval);
void remap_gradients(REMAPGRID rg, double *array, double *grad1_lat,
		     double *grad1_lon, double *grad1_latlon);
