#include <stdio.h>
#include <math.h>

#include "cdo.h"
#include "cdi.h"
#ifndef _DMEMORY_H
#  include "dmemory.h"
#endif

#ifndef _ERROR_H
#  include "error.h"
#endif
#include "specspace.h"

typedef struct {
  double *poli, *pold;
  int trunc, nlat;
}
LegPol;

#define  LEGPOLMAX  5
static LegPol legpol[LEGPOLMAX];
static int nlegpol = 0;

int nlat2trunc(int nlat)
{
  int trunc;

  trunc = (nlat*2 - 1) / 3;

  return (trunc);
}

int nsp2trunc(int nsp)
{
  /*  nsp = (trunc+1)*(trunc+1)              */
  /*      => trunc^2 + 3*trunc - (x-2) = 0   */
  /*                                         */
  /*  with:  y^2 + p*y + q = 0               */
  /*         y = -p/2 +- sqrt((p/2)^2 - q)   */
  /*         p = 3 and q = - (x-2)           */
  int trunc;

  trunc = (int) ((sqrt(nsp*4.0 + 1.0) - 3) / 2);

  return (trunc);
}

int trunc2nlat(int trunc)
{
  int nlat;

  nlat = (trunc*3 + 1) / 2;

  if ( (nlat % 2) > 0 ) nlat++;

  return (nlat);
}


void legini(int truncation, int nlat, int ipol)
{
  static char func[] = "legini";
  int waves, dimsp;
  int jgl, jm, jn;
  int jsp, pdim;
  double *gmu, *gwt, *hnm, *pnm, *ztemp1, *ztemp2;

  waves  = truncation + 1;
  dimsp  = (truncation + 1) * (truncation + 2);
  pdim   = dimsp / 2 * nlat;

  legpol[ipol].trunc  = truncation;
  legpol[ipol].nlat   = nlat;
  legpol[ipol].poli   = (double *) malloc(pdim * sizeof(double));
  legpol[ipol].pold   = (double *) malloc(pdim * sizeof(double));

  gmu    = (double *) malloc(nlat * sizeof(double));
  gwt    = (double *) malloc(nlat * sizeof(double));
  hnm    = (double *) malloc(dimsp * sizeof(double));
  pnm    = (double *) malloc(dimsp * sizeof(double));
  ztemp1 = (double *) malloc((waves<<1) * sizeof(double));
  ztemp2 = (double *) malloc((waves<<1) * sizeof(double));

  gaussaw(gmu, gwt, nlat);

  for ( jgl = 0; jgl < nlat; jgl++ )
    {
      phcs(pnm, hnm, waves, gmu[jgl], ztemp1, ztemp2);
      jsp = jgl;
      for ( jm = 0; jm < waves; jm++ )
	for ( jn = 0; jn < waves - jm; jn++ )
	  {
	    legpol[ipol].poli[jsp] = pnm[jm*waves+jn] * 2.0;
	    legpol[ipol].pold[jsp] = pnm[jm*waves+jn] * gwt[jgl];

	    jsp += nlat;
	  }
    }

  free(ztemp2);
  free(ztemp1);
  free(pnm);
  free(hnm);
  free(gwt);
  free(gmu);
}

double *leg_pold(int trunc, int nlat)
{
  static char func[] = "leg_pold";
  int ipol;

  for ( ipol = 0; ipol < nlegpol; ipol++ )
    if ( legpol[ipol].trunc == trunc && legpol[ipol].nlat == nlat ) break;

  if ( ipol == nlegpol )
    {
      if ( nlegpol == LEGPOLMAX )
	Error(func, "Maximum number of different legpol reached!");
      legini(trunc, nlat, nlegpol++);
    }

  return (legpol[ipol].pold);
}

double *leg_poli(int trunc, int nlat)
{
  static char func[] = "leg_poli";
  int ipol;

  for ( ipol = 0; ipol < nlegpol; ipol++ )
    if ( legpol[ipol].trunc == trunc && legpol[ipol].nlat == nlat ) break;

  if ( ipol == nlegpol )
    {
      if ( nlegpol == LEGPOLMAX )
	Error(func, "Maximum number of different legpol reached!");
      legini(trunc, nlat, nlegpol++);
    }

  return (legpol[ipol].poli);
}

void grid2spec(int gridIDin, double *arrayIn, int gridIDout, double *arrayOut)
{
  static char func[] = "grid2spec";
  int trunc, nlat, nlon, nfc;
  int nlev = 1;
  int waves;
  double *fpwork;
    
  trunc = gridInqTrunc(gridIDout);
  nlon  = gridInqXsize(gridIDin);
  nlat  = gridInqYsize(gridIDin);

  waves = trunc + 1;
  nfc   = waves * 2;

  fpwork = (double *) malloc(nlat*nfc*sizeof(double));

  gp2fc(arrayIn, fpwork, nlat, nlon, nlev, nfc);
  fc2sp(fpwork, arrayOut, leg_pold(trunc, nlat), nlev, nlat, trunc);

/*  free(poli);
   poli = NULL; */
  free(fpwork);
}
	      
void spec2grid(int gridIDin, double *arrayIn, int gridIDout, double *arrayOut)
{
  static char func[] = "spec2grid";
  int trunc, nlat, nlon, nfc;
  int nlev = 1;
  int waves;
  double *fpwork;
    
  trunc = gridInqTrunc(gridIDin);
  nlon  = gridInqXsize(gridIDout);
  nlat  = gridInqYsize(gridIDout);

  waves = trunc + 1;
  nfc   = waves * 2;

  fpwork = (double *) malloc(nlat*nfc*sizeof(double));

  sp2fc(arrayIn, fpwork, leg_poli(trunc, nlat), nlev, nlat, nfc, trunc);
  fc2gp(fpwork, arrayOut, nlat, nlon, nlev, nfc);

/*  free(poli);
   poli = NULL; */
  free(fpwork);
}


void spec2spec(int gridIDin, double *arrayIn, int gridIDout, double *arrayOut)
{
  int truncIn, truncOut;

  truncIn  = gridInqTrunc(gridIDin);
  truncOut = gridInqTrunc(gridIDout);

  sp2sp(arrayIn, truncIn, arrayOut, truncOut);
}


void speccut(int gridIDin, double *arrayIn, double *arrayOut, int waves[])
{
  int trunc;

  trunc  = gridInqTrunc(gridIDin);

  spcut(arrayIn, arrayOut, trunc, waves);
}
